/*
 *      d o u b l e _ c
 *
 *  This contains routines that are specific
 *  to the implementation of doubly linked lists
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include "liblistx.h"


/*====================================================== DOUBLE_PREVIOUS */
static 
node_t  DOUBLE_Previous  (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~
 *------------------------------------------------------------------------*/
{
	_list2h_t doublelist = (_list2h_t)list;
	_list2_t  doublenode = (_list2_t)node;

	LIBASSERT (node != NULL);
    doublenode = doublenode->prev;

    if (doublenode != NULL)
    {
        doublenode = (_list2_t)_LIST_USERNODE(doublelist, doublenode);
    }
    return ((node_t)doublenode);

}


/*===================================================== DOUBLE_AFTER */
static 
int     DOUBLE_After    (list_t list, node_t newnode, node_t oldnode)
/*      ~~~~~~~~~~~~
 *  Add an entry to a doubly linked list
 *  after the given node.  If the given
 *  node is NULL, then isert at the
 *  start of the list.
 *--------------------------------------------------------------------*/
{
	_list2h_t doublelist = (_list2h_t)list;
	_list2_t  doublenewnode = (_list2_t)newnode;
	_list2_t  doubleoldnode = (_list2_t)oldnode;

    LIBDEBUG ("LIST2_Remove",\
           "'next' pointer already set in node",\
            doublenewnode->next == NULL);

    /*
     * Empty list ?
     */
    if (doublelist->nodes == NULL)
    {
        LIBASSERT (doubleoldnode == NULL);
        LIBASSERT (doublelist->last == NULL);

        doublelist->nodes = doublenewnode;
        doublelist->last = doublenewnode;
    }
    else
    {
        LIBASSERT (doublelist->last != NULL);

        if (doubleoldnode == NULL)
        {
            /*
             *  At start of existing list
             */
            doublelist->nodes->prev = doublenewnode;
            doublenewnode->next = doublelist->nodes;
            doublelist->nodes = doublenewnode;
        }
        else
        {
            /*
             *  Into middle of existing list
             */

            if (doubleoldnode->next == NULL)
            {
                /*
                 *  If the old node was the end of
                 *  the list, then the new one is now.
                 */
                LIBASSERT (doublelist->last == doubleoldnode);
                doublelist->last = doublenewnode;
            }
            else
            {
                /*
                 *  If the old node was in the middle of
                 *  the list, then update next node
                 *  previous pointer.
                 */
                doubleoldnode->next->prev = doublenewnode;
            }
           doublenewnode->next = doubleoldnode->next;

            doubleoldnode->next = doublenewnode;
            doublenewnode->prev = doubleoldnode;
        }
    }

    return (LIST_ERROR_NONE);
}


/*======================================================== DOUBLE_REMOVE */
static 
int     DOUBLE_Remove    (list_t list, node_t node)
/*      ~~~~~~~~~~~~~
 *  Remove the specified entry from a doubly linked list.
 *-----------------------------------------------------------------------*/
#define _list    ((_list2h_t)list)
#define _node ((_list2_t)node)
{
    _list2h_t doublelist = (_list2h_t)list;
	_list2_t  doublenode = (_list2_t)node;
    /*
     *  We had better check for a NULL
     *  parameter as this could cause havoc.
     */
    if (doublenode == NULL)
    {
        return LIST_ERROR_BADNODE;
    }
    LIBDEBUG ("LIST2_Remove", \
           "node is not linked in to a list", \
          (doublenode->next != NULL || doublenode->prev != NULL)
        ||(doublenode == doublelist->nodes && doublenode == doublelist->last) )

    /*
     *  If there is a forward pointer, then
     *  change backward link in next node.
     */
    if (doublenode->next != NULL)
    {
        doublenode->next->prev = doublenode->prev;
    }
    else
    {
        /*
         *  No forward link means this should
         *  be the last node in the list.
         */
        LIBASSERT (doublelist->last == doublenode);
        doublelist->last = doublenode->prev;
    }
    /*
     *  If there is a backward pointer, then
     *  change forward link in previous node.
     */
    if (doublenode->prev != NULL)
    {
        doublenode->prev->next = doublenode->next;
    }
    else
    {
        /*
         *  No backward link means this should
         *  be the first node in the list.
         */
        LIBASSERT (doublelist->nodes == doublenode);
        doublelist->nodes = doublenode->next;
    }

    /*
     *  Tidy up and exit.
     */
    doublenode->next = NULL;
    doublenode->prev = NULL;
    return (LIST_ERROR_NONE);
}
#undef _list
#undef _node


/*========================================================================
/*
 *  This is the list of vectored functions
 *  that are used to support this list type.
 *
 *------------------------------------------------------------------------*/

static struct _listmaster master =
        {
        &_LIST_ListCount,
        &_LIST_ListFirst,
        &_LIST_FifoLast,
        &_LIST_ListNext,
        &DOUBLE_Previous,
        &_LIST_CloneList,
        &_LIST_Destroy,
        &_LIST_NewNode,
        &_LIST_ListAdd,
        &DOUBLE_After,
        &DOUBLE_Remove,
        &_LIST_ListFind,
        LIST_TYPE_DOUBLE,
#ifdef DEBUG
        &_LISTdouble,
        "Double Linked List"
#endif /* DEBUG */
        };

/*
 *  This is the default settings for a new
 *  object of this type.
 */
struct _listclass _LISTdouble =
        {
        &master,                    /* Pointer to list master class */
        sizeof(struct _list2h),     /* Size of the list header */
        sizeof(struct _list2),      /* Size of the node header */
        0,                          /* Size of node data */
        0,                          /* Clone count */
        NULL,                       /* User node Initialise */
        NULL,                       /* User node Destroy */
        NULL,                       /* user node Compare */
        };



