/*
 *      E m b e d C o m . c
 *
 *	This file contains functions that are specific to the
 *	handling of embedded lists, and that are suitable for
 *	inclusion into a RLL.
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  10 July 96  DJW   - First version.  Result or re-doing 'embedding' code.
 */

#include <liblist.h>
#include "liblistx.h"

/*============================================================ EMBED_COUNT */
extern
long	_LIST_EmbedCount (list_t list)
/*		~~~~~~~~~~~~~~~~
 *-------------------------------------------------------------------------*/
{
	return (LIST_Count (((_embedh_t)list)->self));
}


/*============================================================ EMBED_FIRST */
extern
node_t	_LIST_EmbedFirst (list_t list)
/*		~~~~~~~~~~~~~~~~
 *-------------------------------------------------------------------------*/
{
	return LIST_First(((_embedh_t)list)->self);
}


/*============================================================= EMBED_LAST */
extern
node_t	_LIST_EmbedLast (list_t list)
/*		~~~~~~~~~~~~~~~
 *-------------------------------------------------------------------------*/
{
	return LIST_Last(((_embedh_t)list)->self);
}


/*============================================================= EMBED_NEXT */
extern
node_t	_LIST_EmbedNext (list_t list, node_t node)
/*		~~~~~~~~~~~~~~~
 *-------------------------------------------------------------------------*/
{
	return LIST_Next(((_embedh_t)list)->self, _LIST_USERNODE(list,node));
}


/*========================================================= EMBED_PREVIOUS */
extern
node_t	_LIST_EmbedPrevious (list_t list, node_t node)
/*		~~~~~~~~~~~~~~~~~~~
 *-------------------------------------------------------------------------*/
{
	return LIST_Previous(((_embedh_t)list)->self, _LIST_USERNODE(list,node));
}


/*============================================================ EMBED_NEW */
extern
node_t  _LIST_EmbedNew (list_t list, va_list ap)
/*      ~~~~~~~~~~~~~~
 *  When we create a new node for an embedded list, we
 *  really only do it at the outer level.
 *-----------------------------------------------------------------------*/
{
    _embedh_t embedlist;
	list_t  selflist;
    node_t  usernode;

    /*
     *  Set up a new node at the outer level
     */
    embedlist = _LIST_EmbedOuter(list);
	selflist = embedlist->self;
    usernode = selflist->listclass->master->nodenew(selflist, ap);
    if (usernode == NULL)
    {
        return NULL;
    }
    LIBASSERT (_LIST_NodeCheck (selflist, usernode));

#ifdef DEBUG
	while (embedlist->inner != NULL)
	{
	    embedlist = embedlist->inner;

		selflist = embedlist->self;
        _LIST_InitNode(selflist, _LIST_REALNODE(selflist,usernode));
        LIBASSERT (_LIST_NodeCheck (selflist, usernode));
	}
#endif /* DEBUG */

    return (usernode);
}


/*============================================================ EMBED_ADD */
extern
int     _LIST_EmbedAdd (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~
 *  If we are going to add a node, then we need to do it
 *  at all levels.
 *------------------------------------------------------------------------*/
{
    _embedh_t    embedlist;
	node_t  usernode;
    int     reply;

    usernode = _LIST_USERNODE(list,node);
    LIBASSERT (_LIST_NodeCheck(list,usernode));
    LIBASSERT (LIST_Index(list,usernode) == 0);

	for ( embedlist = _LIST_EmbedOuter(list), reply = LIST_ERROR_NOTFOUND; 
							embedlist != NULL ;
								  embedlist = embedlist->inner)
	{
		LIBASSERT (_LIST_NodeCheck((list_t)embedlist,usernode));
		LIBASSERT (LIST_Index((list_t)embedlist,usernode) == 0);
		/*
		* Add to list
		*/
		reply = LIST_Add (embedlist->self, usernode);
	    if (reply != LIST_ERROR_NONE)
		{
			/*
			 *	If we get an error (should not happen)
			 *  then we remove from any lists into which
			 *  this node is already inserted.
			 */
			while (embedlist->outer != NULL)
			{
				embedlist = embedlist->outer;
				(void)LIST_Remove(embedlist->self, usernode);
			}
			break;
		}
	}

    return (reply);
}


/*========================================================== EMBED_AFTER */
extern
int     _LIST_EmbedAfter (list_t list, node_t node, node_t oldnode)
/*      ~~~~~~~~~~~~~~~~
 *------------------------------------------------------------------------*/
{
    _embedh_t    embedlist;
	node_t  usernode;
	node_t  useroldnode;
    int     reply;

    usernode = _LIST_USERNODE(list,node);
    LIBASSERT (_LIST_NodeCheck(list,usernode));
    LIBASSERT (LIST_Index(list,usernode) == 0);
	if ((useroldnode = oldnode) != NULL)
	{
		useroldnode = _LIST_USERNODE(list,oldnode);
		LIBASSERT(_LIST_NodeCheck(list,oldnode));
	}

	for (embedlist = _LIST_EmbedOuter(list), reply = LIST_ERROR_NOTFOUND;
							embedlist != NULL ;
									embedlist = embedlist->inner)
	{
		LIBASSERT (_LIST_NodeCheck((list_t)embedlist,usernode));
		LIBASSERT (LIST_Index((list_t)embedlist,usernode) == 0);

		reply = LIST_After (embedlist->self, usernode, useroldnode);
	    if (reply != LIST_ERROR_NONE)
		{
			/*
			 *	If we get an error (should not happen)
			 *  then we remove from any lists into which
			 *  this node is already inserted.
			 */
			for (embedlist = embedlist->outer ;
						embedlist != NULL ; embedlist = embedlist->outer)
			{
				(void)LIST_Remove(embedlist->self, usernode);
			}
			break;
		}
	}

    return (reply);
}


/*========================================================== EMBED_REMOVE */
extern
int     _LIST_EmbedRemove (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~~
 *  Remove from both lists, calling
 *  the list specific remove routine.
 *------------------------------------------------------------------------*/
{
    _embedh_t    embedlist;
	node_t  usernode;
    int     reply;

    usernode = _LIST_USERNODE(list,node);
    LIBASSERT (_LIST_NodeCheck(list,usernode));
    LIBASSERT (LIST_Index(list,usernode) != 0);

	for (embedlist = _LIST_EmbedOuter(list), reply = LIST_ERROR_NOTFOUND;
				embedlist != NULL;
						embedlist = embedlist->inner)
	{
		LIBASSERT (_LIST_NodeCheck(embedlist->self,usernode));
		LIBASSERT (LIST_Index(embedlist->self,usernode) != 0);

		reply = LIST_Remove (embedlist->self, usernode);
		LIBASSERT (reply == LIST_ERROR_NONE);
	} 

    return (reply);
}



/*============================================================ EMBED_FIND */
extern
node_t  _LIST_EmbedFind  (list_t list, 
/*		~~~~~~~~~~~~~~*/ int (*findfunc)(node_t, va_list), 
                         va_list ap)
/*
 *------------------------------------------------------------------------*/
{
	list_t	selflist = ((_embedh_t)list)->self;

	return selflist->listclass->master->find(selflist, findfunc, ap);
}
